@@ -18,11 +18,7 @@
 
 #include <config.h>
 
-#include <ev-document.h>
-#include <ev-document-thumbnails.h>
-#include <ev-async-renderer.h>
-#include <ev-document-factory.h>
-#include <ev-backends-manager.h>
+#include <evince-document.h>
 
 #include <gio/gio.h>
 
@@ -31,6 +27,15 @@
 
 #define THUMBNAIL_SIZE 128
 
+static gint size = THUMBNAIL_SIZE;
+static const gchar **file_arguments;
+
+static const GOptionEntry goption_options[] = {
+	{ "size", 's', 0, G_OPTION_ARG_INT, &size, NULL, "SIZE" },
+	{ G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &file_arguments, NULL, "<input> <ouput>" },
+	{ NULL }
+};
+
 struct AsyncData {
 	EvDocument  *document;
 	const gchar *output;
@@ -65,13 +70,17 @@ evince_thumbnail_pngenc_get (EvDocument *document, const char *thumbnail, int si
 	EvRenderContext *rc;
 	double width, height;
 	GdkPixbuf *pixbuf;
+	EvPage *page;
 
-	ev_document_get_page_size (document, 0, &width, &height);
+	page = ev_document_get_page (document, 0);
+	
+	ev_document_get_page_size (document, page, &width, &height);
 
-	rc = ev_render_context_new (0, 0, size / width);
+	rc = ev_render_context_new (page, 0, size / width);
 	pixbuf = ev_document_thumbnails_get_thumbnail (EV_DOCUMENT_THUMBNAILS (document),
 						       rc, FALSE);
 	g_object_unref (rc);
+	g_object_unref (page);
 	
 	if (pixbuf != NULL) {
 		const char *overlaid_icon_name = NULL;
@@ -127,41 +136,65 @@ evince_thumbnail_pngenc_get_async (struct AsyncData *data)
 	return NULL;
 }
 
+static void
+print_usage (GOptionContext *context)
+{
+	gchar *help;
+
+	help = g_option_context_get_help (context, TRUE, NULL);
+	g_print ("%s", help);
+	g_free (help);
+}
+
 int
 main (int argc, char *argv[])
 {
-	EvDocument *document;
-	const char *input;
-	const char *output;
-	int         size;
-	char       *uri;
-	GFile      *file;
+	EvDocument     *document;
+	GOptionContext *context;
+	const char     *input;
+	const char     *output;
+	char           *uri;
+	GFile          *file;
+	GError         *error = NULL;
+
+	context = g_option_context_new ("- GNOME Document Thumbnailer");
+	g_option_context_add_main_entries (context, goption_options, NULL);
+
+	if (!g_option_context_parse (context, &argc, &argv, &error)) {
+		g_printerr ("%s\n", error->message);
+		g_error_free (error);
+		print_usage (context);
+		g_option_context_free (context);
 
-	if (argc <= 2 || argc > 5 || strcmp (argv[1], "-h") == 0 ||
-	    strcmp (argv[1], "--help") == 0) {
-		g_print ("Usage: %s [-s <size>] <input> <output>\n", argv[0]);
 		return -1;
 	}
 
-	if (!strcmp (argv[1], "-s")) {
-		input = argv[3];
-		output = argv[4];
-		size = atoi (argv[2]);
-	} else {
-		input = argv[1];
-		output = argv[2];
-		size = THUMBNAIL_SIZE;
-	}
+	input = file_arguments ? file_arguments[0] : NULL;
+	output = input ? file_arguments[1] : NULL;
+	if (!input || !output) {
+		print_usage (context);
+		g_option_context_free (context);
 
-	if (size < 40) {
-		g_print ("Size cannot be smaller than 40 pixels\n");
 		return -1;
 	}
+	
+	g_option_context_free (context);
+
+	if (size < 1) {
+		g_print ("Size cannot be smaller than 1 pixel\n");
+		return -1;
+	}
+
+	input = file_arguments[0];
+	output = file_arguments[1];
+
+	g_type_init ();
 
 	if (!g_thread_supported ())
 		g_thread_init (NULL);
 
-	ev_backends_manager_init ();
+        if (!ev_init ())
+                return -1;
 
 	file = g_file_new_for_commandline_arg (input);
 	uri = g_file_get_uri (file);
@@ -171,13 +204,13 @@ main (int argc, char *argv[])
 	g_free (uri);
 
 	if (!document) {
-		ev_backends_manager_shutdown ();
+		ev_shutdown ();
 		return -2;
 	}
 
 	if (!EV_IS_DOCUMENT_THUMBNAILS (document)) {
 		g_object_unref (document);
-		ev_backends_manager_shutdown ();
+		ev_shutdown ();
 		return -2;
 	}
 
@@ -196,19 +229,19 @@ main (int argc, char *argv[])
 		gtk_main ();
 
 		g_object_unref (document);
-		ev_backends_manager_shutdown ();
+		ev_shutdown ();
 
 		return data.success ? 0 : -2;
 	}
 
 	if (!evince_thumbnail_pngenc_get (document, output, size)) {
 		g_object_unref (document);
-		ev_backends_manager_shutdown ();
+		ev_shutdown ();
 		return -2;
 	}
 
 	g_object_unref (document);
-	ev_backends_manager_shutdown ();
+        ev_shutdown ();
 
 	return 0;
 }
